@@ -36,7 +36,7 @@
 
 #define APP_NAME		"eventstat"
 #define TIMER_STATS		"/proc/timer_stats"
-#define TABLE_SIZE		(1997)		/* Should be a prime */
+#define TABLE_SIZE		(32771)		/* Should be a prime */
 
 #define OPT_QUIET		(0x00000001)
 #define OPT_CUMULATIVE		(0x00000002)
@@ -74,7 +74,6 @@ typedef struct timer_info {
 typedef struct timer_stat {
 	unsigned long	count;		/* Number of events */
 	unsigned long	delta;		/* Change in events since last time */
-	bool		old;		/* Existing event, not a new one */
 	timer_info_t	*info;		/* Timer info */
 	struct timer_stat *next;	/* Next timer stat in hash table */
 	struct timer_stat *sorted_freq_next;	/* Next timer stat in event frequency sorted list */
@@ -88,22 +87,127 @@ typedef struct sample_delta_item {
 
 /* list of sample_delta_items */
 typedef struct sample_delta_list {
-	unsigned long		whence;	/* when the sample was taken */
-	list_t			list;
+	struct timeval 	whence;		/* when the sample was taken */
+	list_t		list;
 } sample_delta_list_t;
 
+typedef struct {
+	char *task;			/* Name of kernel task */
+	size_t len;			/* Length */
+} kernel_task_info;
+
+#define KERN_TASK_INFO(str)		{ str, sizeof(str) - 1 }
+
 static list_t timer_info_list;		/* cache list of timer_info */
 static list_t sample_list;		/* list of samples, sorted in sample time order */
 static char *csv_results;		/* results in comma separated values */
 static volatile bool stop_eventstat = false;	/* set by sighandler */
-static unsigned long opt_threshold;	/* ignore samples with event delta less than this */
+static double  opt_threshold;		/* ignore samples with event delta less than this */
 static unsigned int opt_flags;		/* option flags */
+static bool sane_procs;			/* false if we are in a container */
+
+/*
+ *  sane_proc_pid_info()
+ *	detect if proc info mapping from /proc/timer_stats
+ *	maps to proc pids OK. If we are in a container or
+ *	we can't tell, return false.
+ */
+static bool sane_proc_pid_info(void)
+{
+	FILE *fp;
+	static const char pattern[] = "container=";
+	const char *ptr = pattern;
+	bool ret = true;
+
+	fp = fopen("/proc/1/environ", "r");
+	if (!fp)
+		return false;
+
+	while (!feof(fp)) {
+		int ch = getc(fp);
+
+		if (*ptr == ch) {
+			ptr++;
+			/* Match? So we're inside a container */
+			if (*ptr == '\0') {
+				ret = false;
+				break;
+			}
+		} else {
+			/* No match, skip to end of var and restart scan */
+			do {
+				ch = getc(fp);
+			} while ((ch != EOF) && (ch != '\0'));
+			ptr = pattern;
+		}
+	}
+
+	fclose(fp);
+
+	return ret;
+}
+
+/*
+ *  timeval_sub()
+ *	timeval a - b
+ */
+static struct timeval timeval_sub(const struct timeval *a, const struct timeval *b)
+{
+	struct timeval ret, _b;
+
+	_b.tv_sec = b->tv_sec;
+	_b.tv_usec = b->tv_usec;
+
+	if (a->tv_usec < _b.tv_usec) {
+		int nsec = ((_b.tv_usec - a->tv_usec) / 1000000) + 1;
+		_b.tv_sec += nsec;
+		_b.tv_usec -= (1000000 * nsec);
+	}
+	if (a->tv_usec - _b.tv_usec > 1000000) {
+		int nsec = (a->tv_usec - _b.tv_usec) / 1000000;
+		_b.tv_sec -= nsec;
+		_b.tv_usec += (1000000 * nsec);
+	}
+
+	ret.tv_sec = a->tv_sec - _b.tv_sec;
+	ret.tv_usec = a->tv_usec - _b.tv_usec;
+
+	return ret;
+}
+
+/*
+ *  timeval_add()
+ *	timeval a + b
+ */
+static struct timeval timeval_add(const struct timeval *a, const struct timeval *b)
+{
+	struct timeval ret;
+
+	ret.tv_sec = a->tv_sec + b->tv_sec;
+	ret.tv_usec = a->tv_usec + b->tv_usec;
+	if (ret.tv_usec > 1000000) {
+		int nsec = (ret.tv_usec / 1000000);
+		ret.tv_sec += nsec;
+		ret.tv_usec -= (1000000 * nsec);
+	}
+
+	return ret;
+}
+
+/*
+ *  timeval_double
+ *	timeval to a double
+ */
+static inline double timeval_double(const struct timeval *tv)
+{
+	return (double)tv->tv_sec + ((double)tv->tv_usec / 1000000.0);
+}
 
 /*
  *  set_timer_stat()
  *	enable/disable timer stat
  */
-void set_timer_stat(const char *str, const bool carp)
+static void set_timer_stat(const char *str, const bool carp)
 {
 	FILE *fp;
 
@@ -122,13 +226,13 @@ void set_timer_stat(const char *str, const bool carp)
 /*
  *  Stop gcc complaining about no return func
  */
-void eventstat_exit(const int status) __attribute__ ((noreturn));
+static void eventstat_exit(const int status) __attribute__ ((noreturn));
 
 /*
  *  eventstat_exit()
  *	exit and set timer stat to 0
  */
-void eventstat_exit(const int status)
+static void eventstat_exit(const int status)
 {
 	set_timer_stat("0", false);
 
@@ -226,7 +330,7 @@ static void samples_free(void)
  *  sample_add()
  *	add a timer_stat's delta and info field to a list at time position whence
  */
-static void sample_add(timer_stat_t *timer_stat, const unsigned long whence)
+static void sample_add(timer_stat_t *timer_stat, struct timeval *whence)
 {
 	link_t	*link;
 	bool	found = false;
@@ -238,7 +342,8 @@ static void sample_add(timer_stat_t *timer_stat, const unsigned long whence)
 
 	for (link = sample_list.head; link; link = link->next) {
 		sdl = (sample_delta_list_t*)link->data;
-		if (sdl->whence == whence) {
+		if ((sdl->whence.tv_sec == whence->tv_sec) &&
+		    (sdl->whence.tv_usec == whence->tv_usec)) {
 			found = true;
 			break;
 		}
@@ -253,7 +358,7 @@ static void sample_add(timer_stat_t *timer_stat, const unsigned long whence)
 			fprintf(stderr, "Cannot allocate sample delta list\n");
 			eventstat_exit(EXIT_FAILURE);
 		}
-		sdl->whence = whence;
+		sdl->whence = *whence;
 		list_append(&sample_list, sdl);
 	}
 
@@ -300,15 +405,66 @@ static int info_compare_total(const void *item1, const void *item2)
  *  pid_a_kernel_thread
  *
  */
-static bool pid_a_kernel_thread(const pid_t id)
+static bool pid_a_kernel_thread(const char *task, const pid_t id)
 {
 	char buffer[128];
-	char path[32];
+	char path[128];
 
-	snprintf(buffer, sizeof(buffer), "/proc/%d/exe", id);
-	if (readlink(buffer, path, sizeof(path)) < 0)
-		if (errno == ENOENT)
-			return true;
+	if (sane_procs) {
+		int piddir;
+		bool ret = false;
+
+		snprintf(buffer, sizeof(buffer), "/proc/%d", id);
+		if ((piddir = openat(0, buffer, O_RDONLY)) < 0)
+			return ret;	/* it's no more, who knows what it was */
+		
+		if (readlinkat(piddir, "exe", path, sizeof(path)) < 0) 
+			if (errno == ENOENT)
+				ret = true;
+		
+		close(piddir);
+		return ret;
+	} else {
+		/* In side a container, make a guess at kernel threads */
+		int i;
+
+		/*
+		 * This is not exactly accurate, but if we can't look up
+		 * a process then try and infer something from the comm field.
+		 * Until we have better kernel support to map /proc/timer_stats
+		 * pids to containerised pids this is the best we can do.
+		 */
+		static kernel_task_info kernel_tasks[] = {
+			KERN_TASK_INFO("swapper/"),
+			KERN_TASK_INFO("kworker/"),
+			KERN_TASK_INFO("ksoftirqd/"),
+			KERN_TASK_INFO("watchdog/"),
+			KERN_TASK_INFO("migration/"),
+			KERN_TASK_INFO("irq/"),
+			KERN_TASK_INFO("mmcqd/"),
+			KERN_TASK_INFO("jbd2/"),
+			KERN_TASK_INFO("kthreadd"),
+			KERN_TASK_INFO("kthrotld"),
+			KERN_TASK_INFO("kswapd"),
+			KERN_TASK_INFO("ecryptfs-kthrea"),
+			KERN_TASK_INFO("kauditd"),
+			KERN_TASK_INFO("kblockd"),
+			KERN_TASK_INFO("kcryptd"),
+			KERN_TASK_INFO("kdevtmpfs"),
+			KERN_TASK_INFO("khelper"),
+			KERN_TASK_INFO("khubd"),
+			KERN_TASK_INFO("khugepaged"),
+			KERN_TASK_INFO("khungtaskd"),
+			KERN_TASK_INFO("flush-"),
+			KERN_TASK_INFO("bdi-default-"),
+			{ NULL, 0 }
+		};
+
+		for (i = 0; kernel_tasks[i].task != NULL; i++) {
+			if (strncmp(task, kernel_tasks[i].task, kernel_tasks[i].len) == 0)
+				return true;
+		}
+	}
 
 	return false;
 }
@@ -329,9 +485,10 @@ static char *get_pid_cmdline(const pid_t id)
 	if ((fd = open(buffer, O_RDONLY)) < 0)
 		return NULL;
 
-	if ((ret = read(fd, buffer, sizeof(buffer))) <= 0)
+	if ((ret = read(fd, buffer, sizeof(buffer))) <= 0) {
+		close(fd);
 		return NULL;
-
+	}
 	close(fd);
 
 	buffer[sizeof(buffer)-1] = '\0';
@@ -366,7 +523,7 @@ static char *get_pid_cmdline(const pid_t id)
  *  samples_dump()
  *	dump out collected sample information
  */
-static void samples_dump(const char *filename, const int duration)
+static void samples_dump(const char *filename, const struct timeval *duration)
 {
 	sample_delta_list_t	*sdl;
 	timer_info_t **sorted_timer_infos;
@@ -376,6 +533,7 @@ static void samples_dump(const char *filename, const int duration)
 	FILE *fp;
 	unsigned long count = 0;
 	double dur;
+	bool dur_zero;
 
 	if (filename == NULL)
 		return;
@@ -432,18 +590,19 @@ static void samples_dump(const char *filename, const int duration)
 	 *  per sample duration time, instead give the raw sample count
 	 *  by scaling by 1.0 (i.e. no scaling).
 	 */
-	dur = (opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : (double)duration;
+	dur = (opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : timeval_double(duration);
+	dur_zero = (duration->tv_sec == 0) && (duration->tv_usec == 0);
 
 	for (link = sample_list.head; link; link = link->next) {
 		count++;
 		sdl = (sample_delta_list_t*)link->data;
-		fprintf(fp, "%lu", sdl->whence);
+		fprintf(fp, "%f", timeval_double(&sdl->whence));
 
 		/* Scan in timer info order to be consistent for all sdl rows */
 		for (i = 0; i < n; i++) {
 			sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
 			if (sdi)
-				fprintf(fp, ",%f", (double)sdi->delta / dur);
+				fprintf(fp, ",%f", dur_zero ? 0.0 : (double)sdi->delta / dur);
 			else
 				fprintf(fp, ",%f", 0.0);
 		}
@@ -464,7 +623,7 @@ static void samples_dump(const char *filename, const int duration)
 				if (sdi && min > sdi->delta)
 					min = sdi->delta;
 			}
-			fprintf(fp, ",%f", (double)min / dur);
+			fprintf(fp, ",%f", dur_zero ? 0.0 : (double)min / dur);
 		}
 		fprintf(fp, "\n");
 
@@ -478,13 +637,13 @@ static void samples_dump(const char *filename, const int duration)
 				if (sdi && max < sdi->delta)
 					max = sdi->delta;
 			}
-			fprintf(fp, ",%f", (double)max / dur);
+			fprintf(fp, ",%f", dur_zero ? 0.0 : (double)max / dur);
 		}
 		fprintf(fp, "\n");
 
 		fprintf(fp, "Average:");
 		for (i = 0; i < n; i++)
-			fprintf(fp, ",%f", ((double)sorted_timer_infos[i]->total / dur) / (double)count);
+			fprintf(fp, ",%f", dur_zero ? 0.0 : ((double)sorted_timer_infos[i]->total / dur) / (double)count);
 		fprintf(fp, "\n");
 
 		/*
@@ -499,7 +658,7 @@ static void samples_dump(const char *filename, const int duration)
 				sdl = (sample_delta_list_t*)link->data;
 				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
 				if (sdi) {
-					double diff = ((double)sdi->delta - average) / dur;
+					double diff = dur_zero ? 0.0 : ((double)sdi->delta - average) / dur;
 					diff = diff * diff;
 					sum += diff;
 				}
@@ -544,7 +703,7 @@ static timer_info_t *timer_info_find(const timer_info_t *new_info)
 	info->callback = strdup(new_info->callback);
 	info->ident = strdup(new_info->ident);
 	info->kernel_thread = new_info->kernel_thread;
-	info->total = 0;
+	info->total = new_info->total;
 
 	if (info->task == NULL ||
 	    info->func == NULL ||
@@ -671,6 +830,7 @@ static void timer_stat_add(
 	info.callback = callback;
 	info.ident = buf;
 	info.kernel_thread = kernel_thread;
+	info.total = count;
 
 	ts_new->count  = count;
 	ts_new->info = timer_info_find(&info);
@@ -695,9 +855,10 @@ static timer_stat_t *timer_stat_find(
 		needle->info->pid, needle->info->task,
 		needle->info->func, needle->info->callback);
 
-	for (ts = haystack[hash_pjw(buf)]; ts; ts = ts->next)
+	for (ts = haystack[hash_pjw(buf)]; ts; ts = ts->next) {
 		if (strcmp(ts->info->ident, buf) == 0)
 			return ts;
+	}
 
 	return NULL;	/* no success */
 }
@@ -734,9 +895,9 @@ static void timer_stat_sort_freq_add(
  *	silently die
  */
 static void timer_stat_diff(
-	const int duration,		/* time between each sample */
+	struct timeval *duration,	/* time between each sample */
 	const int n_lines,		/* number of lines to output */
-	const unsigned long whence,	/* nth sample */
+	struct timeval *whence,		/* nth sample */
 	timer_stat_t *timer_stats_old[],/* old timer stats samples */
 	timer_stat_t *timer_stats_new[])/* new timer stats samples */
 {
@@ -744,6 +905,7 @@ static void timer_stat_diff(
 	int j = 0;
 	unsigned long total = 0UL;
 	unsigned long kt_total = 0UL;
+	double dur = timeval_double(duration);
 
 	timer_stat_t *sorted = NULL;
 
@@ -756,15 +918,13 @@ static void timer_stat_diff(
 			if (found) {
 				ts->delta = ts->count - found->count;
 				if (ts->delta >= opt_threshold) {
-					ts->old = true;
 					timer_stat_sort_freq_add(&sorted, ts);
 					sample_add(ts, whence);
 					found->info->total += ts->delta;
 				}
 			} else {
-				ts->delta = 0;
+				ts->delta = ts->count;
 				if (ts->delta >= opt_threshold) {
-					ts->old = false;
 					timer_stat_sort_freq_add(&sorted, ts);
 					sample_add(ts, whence);
 				}
@@ -781,15 +941,13 @@ static void timer_stat_diff(
 			if (((n_lines == -1) || (j < n_lines)) && (sorted->delta != 0)) {
 				j++;
 				if (opt_flags & OPT_CUMULATIVE) {
-					printf("%1s %6lu %5d %-15s %-25s %-s\n",
-						sorted->old ? " " : "N",
+					printf("%8lu %5d %-15s %-25s %-s\n",
 						sorted->count,
 						sorted->info->pid, sorted->info->task,
 						sorted->info->func, sorted->info->callback);
 				} else {
-					printf("%1s %6.2f %5d %-15s %-25s %-s\n",
-						sorted->old ? " " : "N",
-						(double)sorted->delta / (double)duration,
+					printf("%8.2f %5d %-15s %-25s %-s\n",
+						(double)sorted->delta / dur,
 						sorted->info->pid, sorted->info->task,
 						sorted->info->func, sorted->info->callback);
 				}
@@ -800,10 +958,13 @@ static void timer_stat_diff(
 
 			sorted = sorted->sorted_freq_next;
 		}
-		printf("%lu Total events, %5.2f events/sec (kernel: %5.2f, userspace: %5.2f)\n\n",
-			total, (double)total / duration,
-			(double)kt_total / duration,
-			(double)(total - kt_total) / duration);
+		printf("%lu Total events, %5.2f events/sec (kernel: %5.2f, userspace: %5.2f)\n",
+			total, (double)total / dur,
+			(double)kt_total / dur,
+			(double)(total - kt_total) / dur);
+		if (!sane_procs)
+			printf("Note: this was run inside a container, kernel tasks were guessed.\n");
+		printf("\n");
 	}
 }
 
@@ -812,7 +973,7 @@ static void timer_stat_diff(
  *	scan /proc/timer_stats and populate a timer stat hash table with
  *	unique events
  */
-void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
+static void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 {
 	FILE *fp;
 	char buf[4096];
@@ -854,9 +1015,14 @@ void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 
 		ptr++;
 		sscanf(buf, "%lu", &count);
-		sscanf(ptr, "%d %s %s (%[^)])", &pid, task, func, timer);
+		if (sscanf(ptr, "%d %s %s (%[^)])", &pid, task, func, timer) != 4)
+			continue;
 
-		kernel_thread = pid_a_kernel_thread(pid);
+		kernel_thread = pid_a_kernel_thread(task, pid);
+
+		/* Swapper is special, like all corner cases */
+		if (strncmp(task, "swapper", 6) == 0)
+			kernel_thread = true;
 
 		if (kernel_thread) {
 			char tmp[64];
@@ -865,19 +1031,10 @@ void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 			strcpy(task, tmp);
 		}
 
-		if (strcmp(task, "swapper") == 0 &&
-		    strcmp(func, "hrtimer_start_range_ns") == 0 &&
-		    strcmp(timer, "tick_sched_timer") == 0) {
-			strcpy(task, "[kern sched]");
-			strcpy(func, "Load balancing tick");
-		}
-
 		if (strcmp(task, "insmod") == 0)
 			strcpy(task, "[kern mod]");
 		if (strcmp(task, "modprobe") == 0)
 			strcpy(task, "[kern mod]");
-		if (strcmp(task, "swapper") == 0)
-			strcpy(task, "[kern core]");
 
 		if ((strncmp(func, "tick_nohz_", 10) == 0) ||
 		    (strncmp(func, "tick_setup_sched_timer", 20) == 0) ||
@@ -894,13 +1051,14 @@ void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
  *  show_usage()
  *	show how to use
  */
-void show_usage(void)
+static void show_usage(void)
 {
 	printf("%s, version %s\n\n", APP_NAME, VERSION);
 	printf("Usage: %s [options] [duration] [count]\n", APP_NAME);
 	printf("Options are:\n");
 	printf("  -c\t\treport cumulative events rather than events per second.\n");
 	printf("  -C\t\treport event count rather than event per second in CSV output.\n");
+	printf("  -d\t\tremove pathname from long process name in CSV output.\n");
 	printf("  -h\t\tprint this help.\n");
 	printf("  -l\t\tuse long cmdline text from /proc/pid/cmdline in CSV output.\n");
 	printf("  -n events\tspecifies number of events to display.\n");
@@ -914,12 +1072,11 @@ void show_usage(void)
 int main(int argc, char **argv)
 {
 	timer_stat_t **timer_stats_old, **timer_stats_new, **tmp;
-	int duration = 1;
+	double duration_secs = 1.0;
 	int count = 1;
 	int n_lines = -1;
-	unsigned long whence = 0;
 	bool forever = true;
-	struct timeval tv1, tv2;
+	struct timeval tv1, tv2, duration, whence;
 
 	list_init(&timer_info_list);
 	list_init(&sample_list);
@@ -974,9 +1131,9 @@ int main(int argc, char **argv)
 	}
 
 	if (optind < argc) {
-		duration = atoi(argv[optind++]);
-		if (duration < 1) {
-			fprintf(stderr, "Duration must be > 0\n");
+		duration_secs = atof(argv[optind++]);
+		if (duration_secs < 0.5) {
+			fprintf(stderr, "Duration must 0.5 or more.\n");
 			eventstat_exit(EXIT_FAILURE);
 		}
 	}
@@ -990,7 +1147,9 @@ int main(int argc, char **argv)
 		}
 	}
 
-	opt_threshold *= duration;
+	duration.tv_sec = (time_t)duration_secs;
+	duration.tv_usec = (suseconds_t)(duration_secs * 1000000.0) - (duration.tv_sec * 1000000);
+	opt_threshold *= duration_secs;
 
 	if (geteuid() != 0) {
 		fprintf(stderr, "%s requires root privileges to write to %s\n",
@@ -998,6 +1157,10 @@ int main(int argc, char **argv)
 		eventstat_exit(EXIT_FAILURE);
 	}
 
+	sane_procs = sane_proc_pid_info();
+	if (!sane_procs)
+		opt_flags &= ~(OPT_CMD_SHORT | OPT_CMD_LONG);
+
 	signal(SIGINT, &handle_sigint);
 
 	if ((timer_stats_old = calloc(TABLE_SIZE, sizeof(timer_stat_t*))) == NULL) {
@@ -1011,24 +1174,42 @@ int main(int argc, char **argv)
 
 	/* Should really catch signals and set back to zero before we die */
 	set_timer_stat("1", true);
-	sleep(1);
-
 	gettimeofday(&tv1, NULL);
 	get_events(timer_stats_old);
 
+	whence.tv_sec = 0;
+	whence.tv_usec = 0;
+
 	while (!stop_eventstat && (forever || count--)) {
-		suseconds_t usec;
+		struct timeval tv;
+		int ret;
 
 		gettimeofday(&tv2, NULL);
-		usec = ((tv1.tv_sec + whence + duration - tv2.tv_sec) * 1000000) +
-		       (tv1.tv_usec - tv2.tv_usec);
-		tv2.tv_sec = usec / 1000000;
-		tv2.tv_usec = usec % 1000000;
 
-		select(0, NULL, NULL, NULL, &tv2);
+		tv = timeval_add(&duration, &whence);
+		tv = timeval_add(&tv, &tv1);
+		tv2 = tv = timeval_sub(&tv, &tv2);
+
+		/* Play catch-up, probably been asleep */
+		if (tv.tv_sec < 0) {
+			tv.tv_sec = 0;
+			tv.tv_usec = 0;
+			tv2 = tv;
+		}
+
+		ret = select(0, NULL, NULL, NULL, &tv2);
+		if (ret < 0) {
+			if (errno == EINTR) {
+				duration = timeval_sub(&tv, &tv2);
+				stop_eventstat = true;
+			} else {
+				fprintf(stderr, "Select failed: %s\n", strerror(errno));
+				break;
+			}
+		}
 
 		get_events(timer_stats_new);
-		timer_stat_diff(duration, n_lines, whence,
+		timer_stat_diff(&duration, n_lines, &whence,
 			timer_stats_old, timer_stats_new);
 		timer_stat_free_contents(timer_stats_old);
 
@@ -1036,10 +1217,10 @@ int main(int argc, char **argv)
 		timer_stats_old = timer_stats_new;
 		timer_stats_new = tmp;
 
-		whence += duration;
+		whence = timeval_add(&duration, &whence);
 	}
 
-	samples_dump(csv_results, duration);
+	samples_dump(csv_results, &duration);
 
 	timer_stat_free_contents(timer_stats_old);
 	timer_stat_free_contents(timer_stats_new);
